package net.mummila.crosswordgame;
import java.io.IOException;
import java.util.Random;

import org.eclipse.swt.widgets.Display;



// a meta class to handle both server and client connections on higher classes' behalf.
public class CGConnection {
	
	// the status reports after attempting to connect to opponent.
	final private static String STATUS_CONNECT_SUCCESS =
		"Yhteys vastapelaajaan muodostettu.";
			
	// 'Waiting for turn' status message.
	final private static String STATUS_WAIT_FOR_TURN = "Odotetaan vastapelaajan siirtoa.";
	
	// the status message used when the server waits for incoming connections.
	final private static String WAITING_FOR_CLIENT = "Odotetaan asiakasyhteytt.";
	
	// whether there is a live connection to an opponent.
	// CGActionButtonSet, CGCrossword, CGLetterSet, CGMenu and CrosswordGame
	// need to know about this.
	public boolean connected;
	
	// the client and the server.
	private CGClient client;
	private CGServer server;
	
    // the constructor sets the parent game and the opponent's IP address,
    // and initializes the connection.
	private CrosswordGame game;
    private String opponentIP;
	public CGConnection(final CrosswordGame g) {
		game = g;
		opponentIP = game.settings.getOpponentIP();
		initialize();
	}
	
	// whether the connection is c2s (TRUE) or a s2c one (FALSE).
	public boolean isClient() {
		return (client != null);
	}
	
	// initialize the connection:
	// kill the server (if any), the client (if any), and go to disconnected state.
	// CrosswordGame needs to know about this.
    public void initialize() {
		if (client != null) {
			client.initialize();
			client = null;
		}
		if (server != null) {
			server.initialize();
			server = null;
		}
    	connected = false;
    }

    // attempt to establish a connection to an opponent.
	public void connect() {	connect(false);	}
	public void connect(boolean continueGame) {

		// initialize any existing connections out of the way.
		initialize();
		
		// disable connecting, enable disconnecting.
		game.actionButtons.connectToDisconnect();
		
		// start the client.
		client = new CGClient(game, opponentIP);

		// if the client connects,
		// set connected status accordingly and exit happily.
		if (client.connect(continueGame)) {
			connected = true;
			game.scoreTable.recordScore(0);
			game.menus.enableQuit();
			game.actionButtons.connectToDisconnect();
			waitForTurn();
			return;
		}
		
		// if the client can't connect, kill it,
		// disable the game, start the server and wait for incoming connections.
		client.initialize();
		client = null;
		
		game.disable();
        game.setStatus(WAITING_FOR_CLIENT);
    	new Thread (new ServerConnect(continueGame)).start();
	}
	
	// send a move to opponent.
	// if sending fails, initializes the connection.
	// returns TRUE with success, otherwise FALSE.
	public boolean sendMove(CGMove move) {
		if (!connected)	return false;
		String sMove = move.encode();
		if (CGMove.getMoveType(sMove).equals(CGProtocol.MOVE_TYPES_GAME_OVER)) {
			if (isClient()) client.sendQuit(sMove);
			else			server.sendQuit(sMove);
			return true;
		}
		else if ( !(isClient() ? client.sendMove(sMove) : server.sendMove(sMove)) ) { 
        	initialize();
        	return false;
		}
		return true;
	}
	
	// execute a given, String-encoded move.
	// returns TRUE if move executed successfully, otherwise returns FALSE.
	public boolean executeMove(String sMove) {
		if (!CGMove.isValidMove(sMove))	return false;
		return new CGMove(game, sMove).execute();
	}

	
	// listen for moves until opponent does a letter swap, a wordfix or a wordfix undo.
	// TODO: or until game ends.
	// CGActionButtonSet and CGMenu need to see this.
	public void waitForTurn() {
		game.disable();
		game.isMyTurn = false;
		final String slots = game.playerSlots.toString();
		game.playerSlots.enable();
		game.setStatus(STATUS_WAIT_FOR_TURN);
		(new Thread (new CGWaiter(game, slots))).start();
	}
	
    // opponent's IP address encapsulation.
    // CGAddressDialog needs these.
    public String getOpponentIP() {
    	return opponentIP;
    }
    public void setOpponentIP(String ip) {
    	opponentIP = ip;
    }
    
    // meta methods for providing I/O to CGWaiter.
    public String readLine() throws IOException {
    	if (client != null)			return client.readLine();
    	else if (server != null)	return server.readLine();
    	else						throw new IOException();
    }
    public void writeLine(String line) throws IOException {
    	if (line == null)			throw new IOException();
    	if (client != null)			client.writeLine(line);
    	else if (server != null)	server.writeLine(line);
    	else						throw new IOException();
    }
    
    
    
    private class ServerConnect implements Runnable {
    	final private boolean continueGame;
    	
    	private ServerConnect(final boolean contnue) {
    		continueGame = contnue;
    		server = new CGServer(game, opponentIP);
    	}
    	
    	public void run() {
    		while (true) {
	    		if (game.actionButtons.connecting) {
	    			
		    		// if the server connects,
		    		// set connected status accordingly, enable the game and exit happily.
	    			try {
			    		if (server.connect(continueGame)) {
			    	    	Display.getDefault().syncExec(new Runnable() {
					    		public void run() {
					    			if (server.sendGame(continueGame)) {
						    			connected = true;
										game.scoreTable.recordScore(0);
						    			game.enable();
										game.menus.enableQuit();
										game.menus.disableUndo();
						    			
						    			// disable 'Swap' if out of alphabet already.
						    			if ( !(game.alphabet.size() > 0) )	game.actionButtons.disableSwap();
						
								        game.setStatus(STATUS_CONNECT_SUCCESS);
								        
								        // pass the first turn to client on a coin flip.
										if ((new Random()).nextBoolean()) {
											CGMove move = new CGMove(game, CGProtocol.MOVE_TYPES_PASS, 0);
											game.connection.sendMove(move);
											game.isMyTurn = false;
											game.connection.waitForTurn();
										}

					    			}
						    	}
					    	});
			    			return;
			    		}
			    		server.initialize();
	    			}
	    			
		    		// the server could be killed by Disconnect.
		    		catch (NullPointerException ex) {
		    			;
		    		}
		    		
		    		// if the server can't connect, try, try again.
	    		}
	    		
	    		// when the player wants to stop trying, stop.
	    		else {
					break;
	    		}
    		}
    	}
    }
}
