package net.mummila.crosswordgame;



// a specialized class for word searching and other advanced algorithms.
public class CGSearcher {
	
	final private CGCrossword crossword;
	final private CrosswordGame game;
	public CGSearcher(final CGCrossword xword, final CrosswordGame g) {
		crossword = xword;
		game = g;
	}
	
	// find the first unfixed letter on the board,
	// counting from top left cell, moving from left to right, top to bottom.
	// returns null if no unfixed cells found.
	private CGCell firstUnfixed() {
		for (int i=0; i<CGCrossword.ROWS; i++) {
			for (int j=0; j<CGCrossword.COLUMNS; j++) {
				CGCell cell = crossword.getCrosswordCell(j, i);
				if (!cell.isEmpty() && !cell.fixed) {
					return cell;
				}
			}
		}
		return null;
	}
	
	// find the first unfixed letter on the board,
	// counting from bottom right cell, moving from right to left, bottom to top.
	// returns null if no unfixed cells found.
	private CGCell lastUnfixed() {
		for (int i=CGCrossword.ROWS-1; !(i<0); i--) {
			for (int j=CGCrossword.COLUMNS-1; !(j<0); j--) {
				CGCell cell = crossword.getCrosswordCell(j, i);
				if (!cell.isEmpty() && !cell.fixed) {
					return cell;
				}
			}
		}
		return null;
	}

	// find the first letter related to a given letter,
	// where the relation is the two being letters of the same word on the same row.
	// returns the starting letter if it is the first letter of the word.
	// returns null if no such relations are to be found on the board.
	private CGCell rowWordStart(final CGCell start) {
		if (start.y > CGCrossword.ROWS)	return null;
		
		int i;
		for (i=start.x-1; !(i<0); i--) {
			if (crossword.getCrosswordCell(i, start.y).isEmpty())	break;
		}
		// i reaching -1 is okay, it just means the word begins at the first column.
		return crossword.getCrosswordCell(i+1, start.y);
	}
	
	// find the first letter related to a given letter,
	// where the relation is the two being letters of the same word on the same column.
	// returns the starting letter if it is the first letter of the word.
	// returns null if no such relations are to be found on the board.
	private CGCell columnWordStart(final CGCell start) {
		if (start.x > CGCrossword.COLUMNS)	return null;
		
		int i;
		for (i=start.y-1; !(i<0); i--) {
			if (crossword.getCrosswordCell(start.x, i).isEmpty())	break;
		}
		
		// i reaching -1 is okay, it just means the word begins at the first row.
		return crossword.getCrosswordCell(start.x, i+1);
	}
	
	// find the last letter related to a given letter,
	// where the relation is the two being letters of the same word on the same row.
	// returns the starting letter if it is the last letter of the word.
	// returns null if no such relations are to be found on the board.
	private CGCell rowWordEnd(final CGCell start) {
		if (start.y > CGCrossword.ROWS)	return null;
		
		int i;
		for (i=start.x+1; i<CGCrossword.COLUMNS; i++) {
			if (crossword.getCrosswordCell(i, start.y).isEmpty())	break;
		}
		
		// i reaching CGCrossword.COLUMNS is okay,
		// it just means the word ends at the last column.
		return crossword.getCrosswordCell(i-1, start.y);
	}
	
	// find the slot of the last letter related to a given letter,
	// where the relation is the two being letters of the same word on the same column.
	// returns the starting letter if it is the last letter of the word.
	// returns null if no such relations are to be found on the board.
	private CGCell columnWordEnd(final CGCell start) {
		if (start.x > CGCrossword.COLUMNS)	return null;
		
		int i;
		for (i=start.y+1; i<CGCrossword.ROWS; i++) {
			if (crossword.getCrosswordCell(start.x, i).isEmpty())	break;
		}
		
		// i reaching CGCrossword.ROWS is okay,
		// it just means the word ends at the last row.
		return crossword.getCrosswordCell(start.x, i-1);
	}
	
	// whether there are empty cells between two given cells on the same row.
	// if the given cells aren't on the same row, the result is indeterminate.
	private boolean rowEmptyBetween(final CGCell cell1, final CGCell cell2) {
		final CGCell start;
		final CGCell end;
		if (cell2.x > cell1.x)	{	start = cell1;	end = cell2;	}
		else					{	start = cell2;	end = cell1;	}
		for (int i=start.x+1; i<end.x; i++) {
			if (crossword.getCrosswordCell(i, start.y).isEmpty())	return true;
		}
		return false;
	}
	
	// whether there are empty cells between two given cells on the same column.
	// if the given cells aren't on the same column, the result is indeterminate.
	private boolean columnEmptyBetween(CGCell cell1, CGCell cell2) {
		final CGCell start;
		final CGCell end;
		if (cell2.y > cell1.y)	{	start = cell1;	end = cell2;	}
		else					{	start = cell2;	end = cell1;	}
		for (int i=start.y+1; i<end.y; i++) {
			if (crossword.getCrosswordCell(start.x, i).isEmpty())	return true;
		}
		return false;
	}
	
	// find the word spanning two given cells.
	// returns null if given letters are unrelated (not on the same row or the same column,
	// or if there are empty cells in between the given letters).
	private CGWord wordContaining(CGCell cell1, CGCell cell2) {
		
		// establish the direction by comparing the cells.
		if (cell1.x == cell2.x) {
			if (columnEmptyBetween(cell1, cell2))	return null;
			CGCell start = columnWordStart(cell1);
			CGCell end = columnWordEnd(cell2);
			return new CGWord(start, end, crossword);
		}
		else if (cell1.y == cell2.y) {
			if (rowEmptyBetween(cell1, cell2))	return null;
			CGCell start = rowWordStart(cell1);
			CGCell end = rowWordEnd(cell2);
			return new CGWord(start, end, crossword);
		}
		else	return null;
		
	}
	
	// find the word containing a given letter.
	// returns null if no such relations are to be found on the board.
	private CGWord wordContaining(CGCell start) {

		if (start == null)	return null;
		
		CGCell verticalNeighbor = null;
		CGCell horizontalNeighbor = null;
		
		boolean isolated = true;

		// while we're at it, seek all fixed neighbors
		// (i.e. don't stop after first one is found).
		// seeking order is TLRB.
		if (start.y > 0) {
			CGCell cell = crossword.getCrosswordCell(start.x, start.y-1);
			if ( !(cell.isEmpty()) ) {
				verticalNeighbor = cell;
				isolated = false;
			}
		}
		if (start.x > 0) {
			CGCell cell = crossword.getCrosswordCell(start.x-1, start.y);
			if ( !(cell.isEmpty()) ) {
				horizontalNeighbor = cell;
				isolated = false;
			}
		}
		if (start.x+1 < CGCrossword.COLUMNS) {
			CGCell cell = crossword.getCrosswordCell(start.x+1, start.y);
			if ( !(cell.isEmpty()) ) {
				horizontalNeighbor = cell;
				isolated = false;
			}
		}
		if (start.y+1 < CGCrossword.ROWS) {
			CGCell cell = crossword.getCrosswordCell(start.x, start.y+1);
			if ( !(cell.isEmpty()) ) {
				verticalNeighbor = cell;
				isolated = false;
			}
		}

		// a bizarre case: a single, isolated letter.
		if (isolated)	return new CGWord(start, start, crossword);
		
		CGWord verticalWord = null;
		CGWord horizontalWord = null;
		if (verticalNeighbor != null) {
			verticalWord = wordContaining(start, verticalNeighbor);
		}
		if (horizontalNeighbor != null) {
			horizontalWord = wordContaining(start, horizontalNeighbor);
		}
		
		if ( (verticalWord == null) && (horizontalWord == null) )	return null;
		if (verticalWord == null)									return horizontalWord;
		else if (horizontalWord == null)							return verticalWord;
		else if (horizontalWord.toString().length() > verticalWord.toString().length())
																	return horizontalWord;
		else														return verticalWord;
		// a vertical word takes preference in cases where the words are equally long.
	}

	// find all unfixed cells between two given cells,
	// return TRUE if they all are in line with the given cells, otherwise return FALSE.
	private boolean allInLineBetween(CGCell firstUnfixed, CGCell lastUnfixed) {

		if (firstUnfixed.y == lastUnfixed.y)	return true;
		if (firstUnfixed.x != lastUnfixed.x)	return false;
		
		for (int j=firstUnfixed.x+1; j<CGCrossword.COLUMNS; j++) {
			CGCell cell = game.crossword.getCrosswordCell(j, firstUnfixed.y); 
			if (!cell.isEmpty() && !cell.fixed)			return false;
		}
		for (int i=firstUnfixed.y+1; i<lastUnfixed.y; i++) {
			for (int j=0; j<CGCrossword.COLUMNS; j++) {
				if (j == firstUnfixed.x)	continue;
				CGCell cell = game.crossword.getCrosswordCell(j, i); 
				if (!cell.isEmpty() && !cell.fixed)		return false;
			}
		}
		for (int j=0; j<lastUnfixed.x; j++) {
			CGCell cell = game.crossword.getCrosswordCell(j, lastUnfixed.y);
			if (!cell.isEmpty() && !cell.fixed)			return false;
		}
		return true;
	}
	
	// get the newly-inserted word.
	// returns an empty String ("") if no new words are on the board.
	// returns an empty String ("") if there are unrelated unfixed letters on the board.
	public String getInsertedWord() {
		
		// find the starting point: the first unfixed letter on the board.
		CGCell firstUnfixed = firstUnfixed();
		
		// if unable to find a starting point, there aren't any new letters on board.
		if ( firstUnfixed == null )	return "";
		
		// find another unfixed letter to establish the word's direction.
		CGCell lastUnfixed = lastUnfixed();

		CGWord word;
		if (lastUnfixed.equals(firstUnfixed))	word = wordContaining(firstUnfixed);
		else if (!allInLineBetween(firstUnfixed, lastUnfixed))	return "";
		else						word = wordContaining(firstUnfixed, lastUnfixed);
		
		// not just a sanity check; the player may try to insert letters erratically.
		if (word == null)	return "";
		else				return word.toString();
	}
	
	//a 'word' is a continuous sequence of CGNewCells inside a CGCrossword.
	private class CGWord {

		final private CGCell start;
		final private CGCell end;
		final private CGCrossword crossword;

		
		
		// the constructor's parameters are the starting cell, the ending cell
		// and the parent crossword.
		public CGWord(final CGCell sInit, final CGCell eInit, final CGCrossword xword) {
			start = sInit;
			end = eInit;
			crossword = xword;
		}

		// whether the word is valid, i.e. whether the starting and ending cell are
		// on the same row or column (return TRUE if they're neither) and whether any
		// of the cell coordinates are outside the crossword's dimensions (return TRUE
		// if any of them are).
		private boolean isInvalid() {
			if ( (start == null) || (end == null) )	return true;
			return ( (start.x != end.x) && (start.y != end.y) );
		}
		
		// return the word as a String.
		// returns null if the word is invalid.
		public String toString() {
			if (isInvalid())	return null;
			
			// establish the direction.
			if (start.x == end.x) {
				String word = "";
				for (int i=start.y; i<end.y+1; i++)
					word = word + crossword.getCrosswordCell(start.x, i).getText();
				
				return word;
			}
			else if (start.y == end.y) {
				String word = "";
				for (int i=start.x; i<end.x+1; i++)
					word = word + crossword.getCrosswordCell(i, start.y).getText();
				
				return word;
			}
			else	return null;
		}
	}
}
