package net.mummila.crosswordgame;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;



// the parent game class.
public class CrosswordGame {
	
	// the main window title.
	final private static String MAIN_TITLE	= "Ristisanapeli";

	// the letterslots' titles.
	final private static String OPPONENT_LETTERS	= "Vastustajan kirjaimet";
	final private static String PLAYER_LETTERS		= "Omat kirjaimet";
	
	// the reset confirmation dialog's contents.
	// which of the two messages is displayed is determined by
	// whether a connection is up
	// when the confirmation is represented.
	// CGMenu needs to know about these.
	final private static String RESET_CONFIRMATION_TITLE = "Varmistus: pelin nollaaminen";
	final private static String RESET_CONFIRMATION_MESSAGE_CONNECTED =
		"Pelin nollaaminen merkitsee sit, ett\n" +
		"- yhteys vastapelaajaan katkaistaan\n" +
		"- pelilauta tyhjennetn kirjaimista\n" +
		"- pelaajille arvotaan uudet kirjaimet.\n" +
		"\n" +
		"Oletko varma ett haluat nollata pelin?";
	final private static String RESET_CONFIRMATION_MESSAGE	=
		"Pelin nollaaminen merkitsee sit, ett\n" +
		"- pelilauta tyhjennetn kirjaimista\n" +
		"- pelaajille arvotaan uudet kirjaimet.\n" +
		"\n" +
		"Oletko varma ett haluat nollata pelin?";
	
	// status messages.
	final private static String INITIALIZED		= "Valmis.";
	final private static String CONNECTION_LOST	= "Yhteys vastapelaajaan katkesi odottamattomasti.";
	final public static String DISCONNECTED		= "Yhteys vastapelaajaan katkaistiin.";

	
	
	// the main window.
	final private Shell shell;

	// the settings.
	// CGMenu needs to see this.
	final public CGSettings settings;

	// the dictionary.
	// CGMenu needs to see this.
	final public CGDictionary dictionary;
	
	// the menus.
	// CGProtocol needs see this.
	final public CGMenu menus;
	
	// game statistics.
	// CGMenu needs to see this.
	final CGStatistics stats = new CGStatistics();
	
	// all the alphabet.
	final public CGAlphabet alphabet;
	
	// the crossword grid, onto which the letters are laid during the game. 
	final public CGCrossword crossword;

	// the letter slots of the local player and the opponent.
	public CGLetterSlots playerSlots;
	public CGLetterSlots opponentSlots;
	
	// the queue for the currently selected letter.
	public CGQueue queue;

	// the action buttons.
	public CGActionButtonSet actionButtons;
	
	// the score table.
	public CGScoreTable scoreTable;
	
	// the status line at the bottom of the window.
	final private Label status;

	// the connection.
	public CGConnection connection = null;

	public boolean isMyTurn;
	
	// whether the game has ended gracefully (due to one of the players choosing to quit).
	public boolean hasEndedGracefully;
	
	
	// the main game constructor.
	private CrosswordGame() {
		final Display display = new Display();
		shell = new Shell(display, SWT.MIN);
		GridLayout gl = new GridLayout();
		shell.setLayout(gl);
		shell.setSize(450, 667);
		shell.setText(MAIN_TITLE);

		// load the settings.
		settings = new CGSettings();
		settings.load(); 
		
		// load the dictionary.
		dictionary = new CGDictionary();
		dictionary.load();

		// initialize the in-game alphabet.
		settings.loadAlphabet();
		alphabet  = new CGAlphabet(this);

		// the menus require that the settings have been initialized.
		menus = new CGMenu(shell, this);

		// once the menu is set up, whip up an exit listener to
		// redirect calls to quit to the 'Exit' menu item's function.
		// http://www.java2s.com/Code/Java/SWT-JFace-Eclipse/Preventashellfromclosingprompttheuser.htm
	    shell.addListener(SWT.Close, new Listener() {
			public void handleEvent(Event event) {
				event.doit = false;
					menus.exit();
			}
	    });
	    
		// visualize the opponent's letters.
		opponentSlots = new CGLetterSlots(OPPONENT_LETTERS, shell, this);
		GridData gd = new GridData(GridData.HORIZONTAL_ALIGN_CENTER);
		opponentSlots.setLayoutData(gd);

		// visualize the crossword grid. 
		crossword = new CGCrossword(shell, this);
		gd = new GridData(GridData.FILL_BOTH);
		crossword.setLayoutData(gd);

		// visualize the local player's letters.
		//playerOldSlots = new CGOldLetterSlots(PLAYER_LETTERS, shell, this);
		playerSlots = new CGLetterSlots(PLAYER_LETTERS, shell, this);
		gd = new GridData(GridData.HORIZONTAL_ALIGN_CENTER);
		playerSlots.setLayoutData(gd);
		
		// visualize the action buttons.
		actionButtons = new CGActionButtonSet(shell, this);
		gd = new GridData(GridData.FILL_HORIZONTAL);

		// visualize the score table.
		scoreTable = new CGScoreTable(shell, this);
		
		// visualize the status bar.
		status = new Label(shell, SWT.CENTER);
		gd = new GridData(GridData.FILL_HORIZONTAL);
		status.setLayoutData(gd);

		// initialize "the game",
		// i.e. the connection, the queue, the alphabet and the visual components.
		initialize();

		shell.pack();
		shell.open();
		while (!shell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}

		finish();
		
		try {
			display.dispose();
		}
		
		// if this happens, we're dead already.
		catch (SWTException ex) {
			System.exit(0);
		}
	}

	// save settings, clean up everything before quitting.
	// TODO: react to failure in saving.
	private void finish() {
		settings.save();
		if (actionButtons.connecting) {
			actionButtons.connecting = false;
			try { Thread.sleep(CGClient.WAIT_MILLISECONDS); }
			catch(InterruptedException ex) { }
		}
		connection.initialize();
	}
	public void exit() {
		dictionary.close();
		shell.dispose();
	}
	
	// initialize the connection.
	private void initializeConnection() {	initializeConnection(false);	}
	private void initializeConnection(final boolean beforeConnecting) {
		if (connection != null) connection.initialize();
		connection = new CGConnection(this);
		if (!beforeConnecting) {
			actionButtons.disconnectToConnect();
			actionButtons.connecting = false;
		}
	}

	// when the connection is lost, some initialization needs to be done.
	// CGActionButtonSet, CGCell, CGLetterSlots and CGWaiter need to see this.
	public void disconnected() {
		disconnected(CONNECTION_LOST);
	}
	public void disconnected(String statusMsg) {
		initializeConnection();
		enable();
		menus.disableUndo();
		menus.disableQuit();
		actionButtons.disconnectToConnect();
		setStatus(statusMsg);
		blink();
	}

	// initialize "the game",
	// i.e. the connection, the queue, the alphabet and the visual components.
	// CGActionButtonSet and CGMenu need to see this.
	public void initialize() {	initialize(false);	}
	public void initialize(final boolean beforeConnecting) {
		
		// initialize the connection.
		initializeConnection(beforeConnecting);

		// initialize the queue for the currently selected letter.
		queue = new CGQueue(this);
		
		// initialize the alphabet.
		alphabet.initialize();
		
		// initialize the letter slot sets.
		opponentSlots.initialize();
		opponentSlots.disable();
		playerSlots.initialize();

		// disable 'Swap' if out of alphabet already.
		if ( !(alphabet.size() > 0) )	actionButtons.disableSwap();
		
		// initialize the crossword grid. 
		crossword.initialize();

		// initialize the score and alphabet counter.
		scoreTable.initialize();

		// disable undo until undo data is available.
		menus.disableUndo();
		menus.disableQuit();
		
		status.setText(INITIALIZED);
		
		hasEndedGracefully = false;
		isMyTurn = true;
	}
	
	// "no moves made" equals "there aren't any letters on the board".
	// CGMenu needs to see this.
	public boolean noMovesMade() {
		return (crossword.countLetters() == 0); 
	}
	
	// a generic confirmation dialog.
	// http://www.java2s.com/Code/Java/SWT-JFace-Eclipse/MessageBoxExample.htm
	public boolean confirmed() {
		return confirmed(
				RESET_CONFIRMATION_TITLE,
				RESET_CONFIRMATION_MESSAGE,
				RESET_CONFIRMATION_MESSAGE_CONNECTED
			);
	}
	public boolean confirmed(final String title, final String message) {
		return confirmed(title, message, message);
	}
	public boolean confirmed(final String title, final String message, final String connectedMessage) {
		MessageBox messageBox = new MessageBox(shell, SWT.ICON_WARNING | SWT.YES | SWT.NO);

		messageBox.setText(title);
		messageBox.setMessage(connection.connected ? connectedMessage : message);
		int selection = messageBox.open();
		boolean confirmation = false;
		switch(selection) {
		case SWT.YES:
			confirmation = true;
			break;
		case SWT.NO:
			confirmation = false;
			break;
		}
		return confirmation;
	}
	
	// start the game.
	public static void main(String[] arg) {	
		new CrosswordGame();
	}

	// status encapsulation.
	public void setStatus(String s) {
		status.setText(s);
	}

	// disable/enable "the game",
	// i.e. disable/enable all the widgets.
	public void disable() {
		menus.disable();
		playerSlots.disable();
		crossword.disable();
		actionButtons.disable();
		scoreTable.disable();
	}
	public void enable() {
		menus.enable();
		playerSlots.enable();
		crossword.enable();
		actionButtons.enable();
		scoreTable.enable();
	}
	
	// set the focus on the window; used to signal the local player of being passed the turn.
	public void blink() {
		shell.forceActive();
	}
}
